// Uebung6 Betriebssysteme 2
// Matthias Jauernig, 2006

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <stdbool.h>
#include <errno.h>
#include <time.h>
#include <signal.h>

#define MAXL 500

#define P_USED 'x'
#define P_UNUSED 'o'

bool ende = false;

void sigintfkt(int signr){
	printf("SIGINT erhalten...\n");
	ende = true;
}

int main(void){
	// Variablen fuer shared memory
	// je 100 Byte fr einen Prozess
	int datalen = (int)(MAXL/5);
	char* data = (char*)malloc(datalen*sizeof(char));
	int shmid, shmflag, shmsize;
	key_t shmkey;
	char* shmadr;
	
	// Variablen fuer Semaphore
	int semid, nsems, semflag;
	key_t key; // Schluessel fuer Semaphormenge
	struct sembuf lock[4], unlock[2];
	
	int sleeptime, i, rc;
	int process = 0; // Nummer unseres Prozesses
	
	// Zufallsgenerator initialisieren
	srand(time(NULL));
	// Signal-Funktion setzen
	sigset(SIGINT, sigintfkt);
	
	// Shared Memory anlegen/oeffnen und Prozess anhaengen
	shmsize = MAXL+5;
	shmkey = ftok("shm_semaphore.c", 1); // Key fuer IPC anhand Filenamen erzeugen
	shmflag = IPC_CREAT | 0644;
	if((shmid=shmget(shmkey, shmsize, shmflag)) < 0){ // Shared Mem erzeugen bzw. oeffnen
		perror("shmget");
		return 1;
	}
	shmadr = shmat(shmid, 0, 0); // Prozess an Shared Memory anhaengen
	
	// prfen, welche Prozessnummer man selber darstellt
	for(i=1; i<=5; i++){
		if(shmadr[MAXL+i-1] != P_USED){  // Prozess noch nicht belegt
			// freier Prozess, trage uns hier ein
			process = i;
			shmadr[MAXL+i-1] = P_USED;
			break;
		}//else: Prozess belegt - weiter suchen
	}	
	if(process == 0)
	{
		printf("Alle 5 Prozesse zuvor gestartet - verlasse...\n");
		return 2;
	}
	printf("Unsere Prozessnummer: %d, Bereiche %d und %d\n", process, process, process%5+1);

	
	// Semaphormenge anlegen
	nsems = 5;	// 5 Semaphore in der Menge
	key = ftok("shm_semaphore.c", 1); // Key anhand Filenamen erzeugen
	semflag = IPC_CREAT | 0666;
	// hole Semaphor-ID anhand des keys
	if((semid = semget(key, nsems, semflag)) < 0){ // wenn Menge schon vorhanden
		if(errno != EEXIST){
			perror("semget");
			shmadr[MAXL+process-1] = P_UNUSED;
			return 3;
		}
		semflag=0666; // versuche auf bestehende Menge zuzugreifen
		if((semid = semget(key,nsems,semflag)) < 0){
        		perror("semget2");
			shmadr[MAXL+process-1] = P_UNUSED;
            		return 4;
        	}
	} // else: Prozess war zuerst da
	
	// Semaphoroperationen eintragen
	// Operationen fr Semaphor (process-1): 1. Bereich dieses Prozesses
	lock[0].sem_num = process-1; // Semaphor-Nr.
	lock[0].sem_op  = 0; // warten, bis Semaphor frei wird
	lock[0].sem_flg = 0;
	lock[1].sem_num = process-1;
	lock[1].sem_op  = 1; // Semaphorwert +1 nach warten
	lock[1].sem_flg = SEM_UNDO; // System kann Operation rueckgaengig machen
	unlock[0].sem_num = process-1;
	unlock[0].sem_op = -1; // nach Freigabe -1
	unlock[0].sem_flg = SEM_UNDO|IPC_NOWAIT; // System kann rueckgaengig machen, kein Warten bei Freigabe
	// Operationen fr Semaphor (process%5): 2. Bereich dieses Prozesses
	lock[2].sem_num = process%5;
	lock[2].sem_op  = 0; // warten, bis Semaphor frei wird
	lock[2].sem_flg = 0;
	lock[3].sem_num = process%5;
	lock[3].sem_op  = 1; // Semaphorwert +1 nach warten
	lock[3].sem_flg = SEM_UNDO;
	unlock[1].sem_num = process%5;
	unlock[1].sem_op = -1; // nach Freigabe -1
	unlock[1].sem_flg = SEM_UNDO|IPC_NOWAIT;
	
	do{
		// warte zufaellige Zeit
		sleeptime = rand()%8;
		printf("Timeout von %d Sekunden...\n", sleeptime);
		sleep(sleeptime);
		if(ende) break;
		
		// fordere beide Bereiche an
		printf("Warte auf Freigabe der Bereiche %d und %d\n", process, process%5+1);
		if((rc=semop(semid,lock,4)) < 0 && !ende){
			// echter Fehler, nicht durch Signal unterbrochen
			perror("semop(lock)");
			shmadr[MAXL+process-1] = P_UNUSED;
			return 5;
		}
		if(ende) break;
		printf("Bereiche gelockt...\n");
		
		// ----------------------
		// <KRITISCHER ABSCHNITT>
			// 1. Bereich des Prozesses auslesen
			strncpy(data,shmadr+(process-1)*datalen,datalen);
	        	if(strlen(data) > 0){ // wenn Daten im shm vorhanden
        		    shmadr[(process-1)*datalen] = '\0'; // Daten im shm loeschen
        		    printf("==> Daten aus Shared Memory Bereich %d: %s\n", process, data);
	        	}
        		else
        		    printf("Keine Daten zum Auslesen aus Bereich %d\n", process);
		        
			// 2. Bereich des Prozesses pruefen
			strncpy(data,shmadr+(process%5)*datalen,datalen); // pruefen, ob Daten in shm vorhanden
       			if(strlen(data) == 0){ // keine Daten vorhanden -> stdin-Eingabe und auf shm schreiben
	       		    printf("Bereich %d ist leer, erwarte Input...\n", process%5+1);
			    printf("Text fuer Bereich %d eingeben: ", process%5+1);
	        	    if(fgets(data, datalen, stdin) != NULL){
				strncpy(shmadr+(process%5)*datalen,data,datalen); // Daten in SharedMemory-Bereich schreiben
		            }
	        	    else{
		                printf("Beende Prozess %d...\n", process);
				shmadr[(process%5)*datalen] = '\0';
	        	        ende = true;
		            }
		        }
	        	else
		            printf("Daten im %d. Bereich vorhanden...\n", process%5+1);	
		// </KRITISCHER ABSCHNITT>
		// -----------------------
		
		printf("Gebe Bereiche %d und %d wieder frei\n", process, process%5+1);
		if((rc=semop(semid,unlock,2)) < 0 && !ende){
			// echter Fehler, nicht durch Signal unterbrochen
			perror("semop(unlock)");
			shmadr[MAXL+process-1] = P_UNUSED;
			return 6;
		}
	}while(!ende);
	
	// Prozess-Datei aus System entfernen
	shmadr[MAXL+process-1] = P_UNUSED;

	// prfen, ob unser Prozess der letzte ist
	process = 0;
	for(i=1; i<=5; i++){
		if(shmadr[MAXL+i-1] == P_USED){ // anderer Prozess existiert noch -> Ende
			process = i;
			break;
		}
		// else: anderer Prozess koennte vll noch existieren -> suche weiter
	}
	
	// Zuordnung des shared mem zum Prozess beenden
	shmdt(shmadr);
	
	// wenn letzter Prozess -> shm und sem-menge aus System entfernen
	if(process == 0){
		printf("Letzter Prozess! Entferne shm-Segment und Semaphormenge...\n");
		shmctl(shmid, IPC_RMID, 0); // shm-Segment aus System entfernen
		semctl(semid, 0, IPC_RMID, 0); // Semaphormenge aus System entfernen
	}
	printf("Beende Programm...\n");
	
	return 0;
}
